# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=missing-docstring
from pytest import mark

from .main import init_logging, main


@mark.parametrize(
    "exp_get, exp_explore",
    (
        # navigating to page fails
        (False, False),
        # exploring page fails
        (True, False),
        # success, result found
        (True, True),
    ),
)
def test_main(mocker, exp_get, exp_explore):
    """test main()"""
    exp = mocker.patch(
        "page_explorer.main.PageExplorer", autospec=True
    ).return_value.__enter__.return_value
    exp.get.return_value = exp_get
    exp.explore.return_value = exp_explore
    assert main(["foo.bin", "http://test.url", "1234"]) == 0
    assert exp.get.call_count == 1
    assert exp.explore.call_count == (1 if exp_get else 0)


@mark.parametrize("level", ("DEBUG", "INFO"))
def test_init_logging(level):
    """test init_logging()"""
    init_logging(level=level)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from contextlib import suppress
from dataclasses import dataclass
from enum import Enum, auto, unique
from logging import getLogger
from os import environ
from time import perf_counter, sleep
from typing import TYPE_CHECKING, Any, Callable, cast

from selenium.common.exceptions import (
    ElementNotInteractableException,
    StaleElementReferenceException,
    WebDriverException,
)
from selenium.webdriver import Firefox as FirefoxDriver
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.firefox.service import Service
from urllib3.exceptions import HTTPError

if TYPE_CHECKING:
    from pathlib import Path

LOG = getLogger(__name__)


@unique
class Action(Enum):
    """Supported actions that can be performed."""

    CLEAR_ELEMENTS = auto()
    EXECUTE_SCRIPT = auto()
    FIND_ELEMENTS = auto()
    KEY_DOWN = auto()
    KEY_UP = auto()
    SEND_KEYS = auto()
    WAIT = auto()


class ExplorerError(Exception):
    """Base exception used by this module."""


@dataclass(frozen=True, eq=False)
class Instruction:
    """Instruction to be executed."""

    action: Action
    delay: float = 0
    runs: int = 1
    value: Any = None


DEFAULT_INSTRUCTIONS = (
    # wait for the page to load more content
    Instruction(Action.WAIT, value=10),
    # Find the end of the page/load more content
    Instruction(Action.SEND_KEYS, value=(Keys.END,), runs=5, delay=0.1),
    Instruction(Action.WAIT, value=1),
    # Attempt to trigger animations
    Instruction(Action.SEND_KEYS, value=(Keys.HOME,)),
    Instruction(Action.SEND_KEYS, value=(Keys.PAGE_DOWN,), runs=10, delay=0.2),
    Instruction(Action.SEND_KEYS, value=(Keys.PAGE_UP,), runs=10, delay=0.1),
    # Select some text
    Instruction(Action.SEND_KEYS, value=(Keys.HOME,)),
    Instruction(Action.KEY_DOWN, value=Keys.SHIFT),
    Instruction(Action.SEND_KEYS, value=(Keys.PAGE_DOWN,)),
    Instruction(Action.KEY_UP, value=Keys.SHIFT),
    Instruction(Action.SEND_KEYS, value=(Keys.HOME,)),
    Instruction(Action.WAIT, value=1),
    # Tab across elements
    Instruction(Action.SEND_KEYS, value=(Keys.TAB,), runs=25),
    Instruction(Action.WAIT, value=1),
    # Zoom in/out
    Instruction(Action.SEND_KEYS, value=(Keys.HOME,)),
    Instruction(
        Action.EXECUTE_SCRIPT,
        value="try { document.body.style.zoom='150%' } catch(e) { }",
    ),
    Instruction(Action.WAIT, value=1),
    Instruction(
        Action.EXECUTE_SCRIPT,
        value="try { document.body.style.zoom='33%' } catch(e) { }",
    ),
    Instruction(Action.WAIT, value=1),
    Instruction(
        Action.EXECUTE_SCRIPT,
        value="try { document.body.style.zoom='100%' } catch(e) { }",
    ),
    Instruction(Action.WAIT, value=1),
    # Find all elements and send ESC
    Instruction(Action.FIND_ELEMENTS, value={"by": By.XPATH, "value": ".//*"}),
    Instruction(Action.SEND_KEYS, value=(Keys.ESCAPE,), runs=25),
    Instruction(Action.CLEAR_ELEMENTS),
    # Call GC (requires fuzzing builds)
    Instruction(
        Action.EXECUTE_SCRIPT,
        value="try { FuzzingFunctions.memoryPressure() } catch(e) { }",
    ),
    Instruction(Action.WAIT, value=1),
)


class PageExplorer:
    """PageExplorer enables page interactions via instructions."""

    __slots__ = ("_driver",)

    def __init__(self, binary: Path, port: int):
        """
        Args:
            binary: Browser binary that is currently running.
            port: Listening browser control port to connect to.
        """
        # disable data collection
        # https://www.selenium.dev/documentation/selenium_manager/#data-collection
        environ["SE_AVOID_STATS"] = "true"
        # Setup the options for connecting to an existing Firefox instance
        options = Options()
        options.binary_location = str(binary)
        options.page_load_strategy = "eager"
        service = Service(
            service_args=[f"--marionette-port={port}", "--connect-existing"],
        )
        try:
            self._driver = FirefoxDriver(options=options, service=service)
        except HTTPError as exc:
            LOG.debug("suppressing HTTPError: %s", exc)
            raise ExplorerError("Failed to create PageExplorer") from None
        except WebDriverException as exc:
            LOG.error("Failed to create WebDriver: %s", exc.msg)
            raise ExplorerError("Failed to create PageExplorer") from None
        LOG.debug("connected to browser on port: %d", port)

    def __enter__(self) -> PageExplorer:
        return self

    def __exit__(self, *exc: object) -> None:
        self.shutdown()

    def close_browser(self, wait: int = 0, poll: float = 0.5) -> None:
        """Attempt to execute 'window.close()' on the browser.

        Args:
            wait: Total amount of time to wait for browser to close.
            poll: Polling interval.

        Returns:
            None.
        """
        assert wait >= 0
        LOG.debug("executing 'window.close()'")
        try:
            self._driver.execute_script(
                "try { window.close() } catch(e) { }"
            )  # type: ignore[no-untyped-call]
        except (HTTPError, WebDriverException):
            LOG.debug("no browser connection")
        else:
            deadline = perf_counter() + wait
            while deadline > perf_counter() and self.is_connected():
                sleep(poll)

    @property
    def current_url(self) -> str | None:
        """Get the URL of the current page.

        Args:
            None

        Returns:
            The URL if it is available otherwise None.
        """
        with suppress(HTTPError, WebDriverException):
            return self._driver.current_url
        return None

    # pylint: disable=too-many-branches
    def explore(
        self,
        instructions: tuple[Instruction, ...] = DEFAULT_INSTRUCTIONS,
        wait_cb: Callable[[float], None] = sleep,
    ) -> bool:
        """Interact with active page by executing provided instructions.

        Args:
            instructions: Instructions to perform.
            wait_cb: Function used to delay execution.

        Returns:
            True if all instructions were successfully executed otherwise False.
        """
        idx = -1
        success = False
        elements = None

        LOG.debug("explore (instructions: %d)", len(instructions))
        # pylint: disable=too-many-nested-blocks
        try:
            actions = ActionChains(self._driver)
            for instruction in instructions:
                idx += 1
                if instruction.action == Action.CLEAR_ELEMENTS:
                    elements = None
                elif instruction.action == Action.EXECUTE_SCRIPT:
                    self._driver.execute_script(
                        instruction.value
                    )  # type: ignore[no-untyped-call]
                elif instruction.action == Action.FIND_ELEMENTS:
                    elements = self._driver.find_elements(
                        **cast(dict[str, str], instruction.value)
                    )
                    if not elements:
                        LOG.debug("no elements found!")
                elif instruction.action == Action.KEY_DOWN:
                    actions.key_down(cast(str, instruction.value)).perform()
                elif instruction.action == Action.KEY_UP:
                    actions.key_up(cast(str, instruction.value)).perform()
                elif instruction.action == Action.SEND_KEYS:
                    if elements is not None:
                        for element in elements:
                            with suppress(
                                ElementNotInteractableException,
                                StaleElementReferenceException,
                            ):
                                element.send_keys(
                                    *cast(tuple[str, ...], instruction.value)
                                )
                            if instruction.delay > 0:
                                wait_cb(instruction.delay)
                    else:
                        for _ in range(instruction.runs):
                            actions.send_keys(
                                *cast(tuple[str, ...], instruction.value)
                            ).perform()
                            if instruction.delay > 0:
                                wait_cb(instruction.delay)
                elif instruction.action == Action.WAIT:
                    wait_cb(cast(float, instruction.value))
            # all instructions complete
            success = True
        except HTTPError as exc:
            LOG.debug("explore - HTTPError: %s", exc)
        except WebDriverException as exc:
            LOG.debug("explore - WebDriverException: %s", exc.msg)
        finally:
            LOG.debug("%d/%d instructions executed", idx + 1, len(instructions))
        return success

    def get(self, url: str) -> bool:
        """Attempt to navigate to a provided URL.

        Args:
            url: URL to load.

        Returns:
            True if URL is successfully loaded otherwise False.
        """
        try:
            self._driver.get(url)
        except HTTPError as exc:
            LOG.debug("get - HTTPError: %s", exc)
        except WebDriverException as exc:
            LOG.debug("get - WebDriverException: %s", exc.msg)
        else:
            LOG.debug("load event received")
            return True
        return False

    def is_connected(self) -> bool:
        """Check if a page is open and connection is active.

        Args:
            None

        Returns:
            True if a page is open and connection is active otherwise False.
        """
        with suppress(HTTPError, WebDriverException):
            return isinstance(self._driver.title, str)
        return False

    def shutdown(self) -> None:
        """Shutdown driver.

        Args:
            None

        Returns:
            None.
        """
        with suppress(HTTPError, WebDriverException):
            self._driver.quit()

    @property
    def title(self) -> str | None:
        """Get the title of the current page.

        Args:
            None

        Returns:
            The page title if it is available otherwise None.
        """
        with suppress(HTTPError, WebDriverException):
            return self._driver.title
        return None



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=missing-docstring
from itertools import count, repeat
from unittest.mock import Mock

from pytest import mark, raises
from selenium.common.exceptions import WebDriverException
from urllib3.exceptions import HTTPError

from .page_explorer import (
    DEFAULT_INSTRUCTIONS,
    Action,
    ExplorerError,
    Instruction,
    PageExplorer,
)


@mark.parametrize("exc", (WebDriverException("test"), HTTPError()))
def test_page_explorer_create(mocker, exc):
    """test creating a PageExplorer object"""
    driver = mocker.patch("page_explorer.page_explorer.FirefoxDriver", autospec=True)

    # create a PageExplorer object
    with PageExplorer("bin", 1234):
        pass
    assert driver.return_value.quit.call_count == 1
    driver.reset_mock()

    # attempt create a PageExplorer object but fail to connect
    driver.side_effect = (exc,)
    with (
        raises(ExplorerError, match="Failed to create PageExplorer"),
        PageExplorer("bin", 1234),
    ):
        pass


@mark.parametrize(
    "title_effect, result",
    (
        # connected
        (("foo",), True),
        # not connected
        ((WebDriverException("test"),), False),
    ),
)
def test_page_explorer_is_connected(mocker, title_effect, result):
    """test PageExplorer.is_connected()"""
    driver = mocker.patch("page_explorer.page_explorer.FirefoxDriver", autospec=True)
    type(driver.return_value).title = mocker.PropertyMock(side_effect=title_effect)

    with PageExplorer("bin", 1234) as exp:
        assert exp.is_connected() == result


@mark.parametrize(
    "url_effect, result",
    (
        # connected
        (("foo",), "foo"),
        # not connected
        ((WebDriverException("test"),), None),
    ),
)
def test_page_explorer_current_url(mocker, url_effect, result):
    """test PageExplorer.current_url"""
    driver = mocker.patch("page_explorer.page_explorer.FirefoxDriver", autospec=True)
    type(driver.return_value).current_url = mocker.PropertyMock(side_effect=url_effect)

    with PageExplorer("bin", 1234) as exp:
        assert exp.current_url == result


@mark.parametrize(
    "title_calls, title_effect, script_effect",
    (
        # wait until deadline is exceeded
        (4, repeat("foo"), None),
        # successfully close the browser
        (1, (WebDriverException("test"),), None),
        # failed to send window.close()
        (0, (AssertionError("test failed"),), (WebDriverException("test"),)),
    ),
)
def test_page_explorer_close_browser(mocker, title_calls, title_effect, script_effect):
    """test PageExplorer.close_browser()"""
    mocker.patch("page_explorer.page_explorer.perf_counter", side_effect=count())
    mocker.patch("page_explorer.page_explorer.sleep", autospec=True)
    driver = mocker.patch(
        "page_explorer.page_explorer.FirefoxDriver", autospec=True
    ).return_value

    fake_title = mocker.PropertyMock(side_effect=title_effect)
    type(driver).title = fake_title
    driver.execute_script.side_effect = script_effect
    with PageExplorer("bin", 1234) as exp:
        exp.close_browser(wait=5)
    assert driver.execute_script.call_count == 1
    assert fake_title.call_count == title_calls


@mark.parametrize(
    "get_effect, expected",
    (
        # successfully get a url
        (None, True),
        # get non-existing url
        ((WebDriverException("test"),), False),
        # browser connection error
        ((HTTPError(),), False),
    ),
)
def test_page_explorer_get(mocker, get_effect, expected):
    """test PageExplorer.get()"""
    driver = mocker.patch(
        "page_explorer.page_explorer.FirefoxDriver", autospec=True
    ).return_value

    driver.get.side_effect = get_effect
    with PageExplorer("bin", 1234) as exp:
        result = exp.get("http://foo.com")
    assert driver.get.call_count == 1
    assert result == expected


@mark.parametrize(
    "instructions, found_elements, expected",
    (
        # one instruction, multiple run
        (
            (Instruction(Action.SEND_KEYS, value=("m",), runs=10, delay=0.1),),
            None,
            True,
        ),
        # send multiple instructions
        (
            (
                Instruction(Action.SEND_KEYS, value=("A",)),
                Instruction(Action.SEND_KEYS, value=("B",)),
            ),
            None,
            True,
        ),
        # send key up and down
        (
            (
                Instruction(Action.KEY_DOWN, value="A"),
                Instruction(Action.KEY_UP, value="A"),
            ),
            None,
            True,
        ),
        # execute script
        ((Instruction(Action.EXECUTE_SCRIPT, value="foo()"),), None, True),
        # wait
        ((Instruction(Action.WAIT, value=1.0),), None, True),
        # find elements - no elements
        (
            (
                Instruction(
                    Action.FIND_ELEMENTS, value={"by": "xpath", "value": ".//*"}
                ),
            ),
            ([],),
            True,
        ),
        # find elements, send key to elements and clear elements
        (
            (
                Instruction(
                    Action.FIND_ELEMENTS, value={"by": "xpath", "value": ".//*"}
                ),
                Instruction(Action.SEND_KEYS, value=("a"), delay=0.1),
                Instruction(Action.CLEAR_ELEMENTS),
            ),
            ([Mock()],),
            True,
        ),
        # find elements, browser closed
        (
            (Instruction(Action.FIND_ELEMENTS, value={}),),
            (WebDriverException("test"),),
            False,
        ),
        # browser connection failed
        ((Instruction(Action.FIND_ELEMENTS, value={}),), (HTTPError(),), False),
        # DEFAULT_INSTRUCTIONS
        (DEFAULT_INSTRUCTIONS, ([Mock()],), True),
    ),
)
def test_page_explorer_explore(mocker, instructions, found_elements, expected):
    """test PageExplorer.explore()"""
    mocker.patch("page_explorer.page_explorer.ActionChains", autospec=True)
    driver = mocker.patch(
        "page_explorer.page_explorer.FirefoxDriver", autospec=True
    ).return_value
    driver.find_elements.side_effect = found_elements

    with PageExplorer("bin", 1234) as exp:
        result = exp.explore(instructions=instructions, wait_cb=mocker.MagicMock())
    assert result == expected


@mark.parametrize(
    "title_effect, result",
    (
        # connected
        (("foo",), "foo"),
        # not connected
        ((WebDriverException("test"),), None),
    ),
)
def test_page_explorer_title(mocker, title_effect, result):
    """test PageExplorer.title"""
    driver = mocker.patch("page_explorer.page_explorer.FirefoxDriver", autospec=True)
    type(driver.return_value).title = mocker.PropertyMock(side_effect=title_effect)

    with PageExplorer("bin", 1234) as exp:
        assert exp.title == result



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from logging import DEBUG, ERROR, INFO, WARNING, basicConfig, getLogger
from time import perf_counter

from .args import parse_args
from .page_explorer import PageExplorer

LOG = getLogger(__name__)
getLogger("selenium").setLevel(WARNING)
getLogger("urllib3").setLevel(ERROR)


def init_logging(level: str) -> None:
    """Initialize logging.

    Arguments:
        level: logging verbosity level

    Returns:
        None
    """
    mapping = {"DEBUG": DEBUG, "ERROR": ERROR, "INFO": INFO, "WARNING": WARNING}
    assert level in mapping
    if level == "DEBUG":
        log_fmt = "%(asctime)s.%(msecs)03d %(levelname).1s %(name)s | %(message)s"
    else:
        log_fmt = "[%(asctime)s] %(message)s"
    basicConfig(format=log_fmt, datefmt="%H:%M:%S", level=mapping[level])


def main(argv: list[str] | None = None) -> int:
    """Main entry point."""
    args = parse_args(argv)
    init_logging(args.log_level)

    try:
        LOG.info("Starting Page Explorer...")
        with PageExplorer(args.binary.resolve(), port=args.port) as explorer:
            LOG.info("Loading %r...", args.url)
            start = perf_counter()
            success = explorer.get(args.url)
            LOG.info("Done. (%0.1fs)", perf_counter() - start)
            if not success:
                LOG.info("Load failed!")
                if explorer.is_connected():
                    LOG.info("Server not found.")
            else:
                LOG.info("Exploring...")
                start = perf_counter()
                success = explorer.explore()
                LOG.info("Done. (%0.1fs)", perf_counter() - start)
                if not success:
                    LOG.info("Could not complete exploration!")
                if explorer.is_connected():
                    LOG.info("Closing browser...")
                    explorer.close_browser(wait=5)

    except KeyboardInterrupt:  # pragma: no cover
        LOG.warning("Aborting...")

    finally:
        LOG.warning("Shutting down...")
    LOG.info("Done.")

    return 0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

try:
    __version__ = version("page-explorer")
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    __version__ = "unknown"


def parse_args(argv: list[str] | None = None) -> Namespace:
    """Argument parsing"""
    parser = ArgumentParser(
        prog="page-explorer", description="Interact with web content."
    )
    parser.add_argument("binary", type=Path, help="Firefox binary in use.")
    parser.add_argument("url", help="URL to visit.")
    parser.add_argument("port", type=int, help="Browser listening port.")

    parser.add_argument(
        "--log-level",
        choices=("DEBUG", "ERROR", "INFO", "WARNING"),
        default="INFO",
        help="Configure console output (default: %(default)s).",
    )

    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=f"%(prog)s {__version__}",
        help="Show version number.",
    )

    return parser.parse_args(argv)
